#!/bin/bash
 
# squish.cron.sh  -- Quick and DIRTY squid log analysis
#
# Assemble squid logs, and send through squish.pl.  This script gets run on a
# regular basis.  It will install itself in crontab if you give it a chance.
#
# Copyright (c) 2001-2002 Andrew McGill and Leading Edge Business Solutions
# (South Africa).  This software may be redistributed and/or modified only
# under the terms of the GNU General Public Licence, version 2, as published by
# the Free Software Foundation, and contained in the file COPYING. 
# 
# All other rights are reserved, and no warranty is implied.  Despite any
# appearance to the contrary, this software is not fit for ANY purpose you may
# conceive, and should not be used on your system under any circumstances.

REVISION='$Id: squish.cron.sh,v 1.3 2005/11/21 15:52:40 andrewm Exp $'
SHORTREVISION="${REVISION/*,v }"
SHORTREVISION="${SHORTREVISION/ *}"

# You can override automatic 'detection' by editing this, or by setting
# an environmnet variable:
# BASEDIR="/usr/local/httpd/htdocs/squish"
# LOGDIR="/var/log/squid"
SQUINT="${SQUINT:=nice /usr/local/bin/squish.pl}"
# If you run setuphtaccess
HTTPDCONF=/etc/httpd/httpd.conf

# Irix doesn't have -maxdepth 1 on find, so we'll work around it
FINDOPTS="-maxdepth 1"
find /dev/null $FINDOPTS >& /dev/null | FINDOPTS=

# Find a nice place to put the output - semi-automatic nonsense that is bound
# to fail sooner or later. If you don't like this, you can change BASEDIR and
# LOGDIR above
[ "$LOGDIR" ] ||
for DIR in \
	/var/squid/logs \
	/var/log/squid \
	/usr/local/squid/var/logs \
	; do
	if [ -d $DIR ] ; then LOGDIR=$DIR; fi
done
[ "$LOGDIR" ] || { echo 1>&2 "Please set LOGDIR in $0"; exit 1; }
umask 002
 
# If we are not on a tty, then log to file
tty -s || {
	exec >& $BASEDIR/squish.log
	date
	trap "date" EXIT
}


function checkcrontab()
{
# install CRONTAB if crontab is writable
CRONTAB="/etc/crontab"
[ -w $CRONTAB ] || return

grep -q "squish" $CRONTAB ||
echo "# squish squishy squid
*/5 * * * * root /usr/local/bin/squish.cron.sh 
" >> $CRONTAB && killall -1 cron crond 2>/dev/null
echo "Modified /etc/crontab"
}

# Find the access logs, and concatenate them to stdout in the right order
# (ie. in the order in which the events happened)
function cataccesslogs()
{
	if [ $ALL ] ; then
		DAYS=365 # older than a year is off the graph
		find $LOGDIR $FINDOPTS -name "access.log*" -mtime -$DAYS |
			xargs ls -tr /dev/null 
	else
		find $LOGDIR $FINDOPTS -name "access.log" 
	fi |
	while read LOGFILE ; do
		[ -r "$LOGFILE" ] && 
		case "$LOGFILE" in 
		*.gz)	zcat "$LOGFILE" ;;
		*.bz2)	bzcat "$LOGFILE" ;;
		*)      cat "$LOGFILE" ;;
		esac
	done
}

function doreport()
{
	# Unzip files before use ...
	# find $LOGDIR $FINDOPTS -name "access.log-*.gz" -mtime -$DAYS -exec gunzip {} \;
	DIR=`dirname $0`
	cd $DIR
	cataccesslogs | $DIR/squish.pl
}

# options processing ...
ALL=
case "$1" in
	--install) 
		ALL=1
		;;
esac
		
doreport

exit 0

