#!/usr/bin/perl -w
#
# squish.pl  --  See who's using resources and squish them
#
# Copyright (c) 2001-2002 Andrew McGill and Leading Edge Business Solutions
# (South Africa).  This software may be redistributed and/or modified only
# under the terms of the GNU General Public Licence, version 2, as published by
# the Free Software Foundation, and contained in the file COPYING. 
# 
# All other rights are reserved, and no warranty is implied.  Despite any
# appearance to the contrary, this software is not fit for ANY purpose you may
# conceive, and should not be used on your system under any circumstances.
# 
# This program reads a squid log from stdin and creates a time and data based
# blocklist for use with squid

use English;

$revision = '$Id: squish.pl,v 1.29 2005/11/21 13:23:08 andrewm Exp $';

no warnings qw{uninitialized};
use POSIX qw(strftime mktime);

use rrdsquish;
do "squish.pm";
# use Storable qw( lock_store lock_retrieve );

$notime = 0;
$conffile= "/etc/squid/squish.conf";
$blacklistfile = "/etc/squid/squished";
(-r $conffile) or $conffile= "squish.conf";
(-w $blacklistfile) or $blacklistfile = "squished";
$reloadcommand = "killall -1 squid";

@excludelist = (
	{ "field" => 3, "pattern" => "TCP_DENIED/" } ,
	{ "field" => 3, "pattern" => "NONE/" },
	{ "field" => 6, "pattern" => '^http://127\.0\.0\.1/' } # localhost
);

%exclude_fieldnames = (
	"date" =>      	"0", "duration" =>	"1", 
	"ip" =>		"2", "status" =>	"3", 
	"bytes" =>	"4", "operation" =>	"5", 
	"url" =>	"6", "user" =>	"7", 
	"method" =>	"8", "content-type" => "9" );

$lines_squished = 0;

getcurrentstate();

# Get the user RRD for bytes or time
sub userrrd($$)
{
	my $user = shift;
	my $type = shift;
	my $rrd = $$userdb{$user}{$type};
	if (!defined $rrd) {
		$rrd = rrdsquish::new();
		$$userdb{$user}{$type} = $rrd;
		if ($type eq "time") {
			$rrd->setmaxperhour(3600);
		}
	}
	#print "RRD for $user -- $type\n";
	#$rrd->print(60,600);
	#sleep 2;
	return $rrd;
}
sub usage()
{
	die "Usage: $0 [options] \n" .
	"\n".
	"Options:\n".
	"  --config <file>  Configuration file (user limits)\n".
	"  --squish <file>  Squished users listed here\n".
	"  --reload \"cmd\"   Command to make squishes take effect\n".
	"Installation and debug\n".
	"  --install        Hack httpd.conf and squid.conf\n".
	"  --notime         Don't use the clock\n".
	# "  --nodns          Sanity check assuming squid.conf says 'log_fqdn off'\n".
	"\n".
	"squish updates a list of people who use\n".
	"over than their quota of time or data on a\n".
	"squid proxy\n".
	"\n";
}

sub showuser($)
{
	my $user=shift;
	my $rrd_time = userrrd($user,"time");
	my $rrd_data = userrrd($user,"data");

	my $timetoascii = strftime ("%a %d %b %Y %H:%M", localtime($rrd_time->{timestamp}));
	printf STDOUT "%-15s $timetoascii  %5d %9d\n",
		$user,
		$rrd_time->total(0),
		$rrd_data->total(0);

	#print $rrd_time->print(0,3600*24*2);

	return;

	print STDOUT "$user: time (m:10m:h:d) ".
		$rrd_time->normalise(60,0,59).":".
		$rrd_time->normalise(60,0,600-1).":".
		$rrd_time->normalise(60,0,3600-1).":".
		$rrd_time->normalise(60,0,86400-1).
		$rrd_time->normalise(60,0,86400*30-1).
		" data (m:5m:h:d) ".
		$rrd_data->sum(60).":".
		$rrd_data->sum(600).":".
		$rrd_data->sum(3600).":".
		$rrd_data->sum(86400)."\n";
	print $rrd_time->print(60,3600*24*2);
	# 250ms delay
	select(undef, undef, undef, 0.25);
}


sub showdata()
{
	print "Name/IP         Last access             time      data\n";
	foreach $user (sort keys %$userdb) {
		showuser($user);
	}
}

sub writekillfile()
{
	my $blacklist;
	{
	open BLACKLIST, "< $blacklistfile";  # ignore errors here ...
	local $/;           # enable "slurp" mode
	$blacklist=<BLACKLIST>;  # ignore errors here ...
	close BLACKLIST;  # ignore errors here ...
	}

	if (! open CONFFILE,"< $conffile" ) {
		print STDERR "$0: $conffile: $!\n";
		return;
	}

	my %squishy;
	$squishy{"aardvarkscantsquishants"}=1;
	foreach my $user (keys %$userdb) {
		$$userdb{$user}{"reason"}='';
	}

	while (<CONFFILE>) {
		my $line = $.;
		# [/etc/squid/overquota]
		chomp;
		s/#.*//;
		s/^\s+//;
		s/\s+$//;
		if (m/^$/) { next; }
		my @line = split /\s+/;
		my $upat=shift @line;
		for my $quota (@line) {
			if ($quota =~ m/(\d+)(h|s|m|mb|kb|b|gb)\/(d|day|w|week|m|month|y|yr|year)/i) {
				my $n=$1;
				my $unit=$2;
				my $period=$3;

				if ($period =~ m/^d/i) { $tperiod=0; }
				elsif ($period =~ m/^w/i) { $tperiod=1; }
				elsif ($period =~ m/^m/i) { $tperiod=2; }
				elsif ($period =~ m/^y/i) { $tperiod=3; }
				else { $tperiod = 0; }
				my $time="t:data";
				my $f=1;
				if ($unit =~ m/^s$/) { $time="t:time"; }
				elsif ($unit =~ m/^m$/) { $f = 60; $time="t:time"; }
				elsif ($unit =~ m/^h$/i) { $f = 3600; $time="t:time"; }
				elsif ($unit =~ m/^b$/i) { $f = 1; } # cool
				elsif ($unit =~ m/^kb$/i) { $f = 1024; }
				elsif ($unit =~ m/^mb$/i) { $f = 1024*1024; }
				elsif ($unit =~ m/^gb$/i) { $f = 1024*1024*1024; }
				foreach my $user (keys %$userdb) {
					if (! ( $user =~ m/^$upat$/ )) { next; }
					my $y = $$userdb{$user}{$time}[$tperiod];
					if ($y > $n*$f) {
						my $reason=sprintf "%.2f$unit is over $n$unit/$period",$y/$f;;
						$squishy{$user}=$reason;
						$$userdb{$user}{"reason"}=$reason;
					}
				}
			}
			else {
				print STDERR "$0: $conffile:$line says \"$quota\" (user \"$upat\")\n";
			};
		}
	}
	close CONFFILE;

	# Write if there was a modification
	my $newblacklist = join ("\n", sort keys %squishy) . "\n";
	if ($newblacklist ne $blacklist) {
		if (! open BLACKLIST, "> $blacklistfile" ) {
			printf STDERR "$0: writing $blacklistfile: $!\n";
			return;
		}
		print BLACKLIST $newblacklist;
		close BLACKLIST;
		print qx{ $reloadcommand };
	}
}

# Some configuration file hacking
sub installit()
{
	my $file;
	my $conf;
	# apache: include
	$file=glob("/etc/{apache,http}*/httpd.conf");
	open CONF,"< $file" || die "$0: can't read $file: $!\n";
	local $/;           # enable "slurp" mode
	$conf=<CONF>;
	close CONF;
	$pwd = $0;
	$pwd =~ s{/[^/]+$}{};
	if ( $conf =~ m/added by squish/ ) {
		print "Already modified $file\n";
	}
	else {
		$conf .= "# added by squish\n".
			"Include $pwd/apache-squish.conf\n";
		print "Writing to $file\n";
		open CONF,"> $file";
		print CONF $conf;
		close CONF;
	}

	# squid
	$file=glob("/etc/squid/squid.conf");
	open CONF,"< $file" || die "$0: can't read $file: $!\n";
	local $/;           # enable "slurp" mode
	$conf=<CONF>;
	close CONF;
	if ( $conf =~ m/added by squish/ ) {
		print "Already modified $file\n";
	}
	else {
		my $hostname = qx(hostname -f);
		chomp $hostname;
		if ($hostname eq '') { $hostname='localhost'; }
		$pwd = $0;
		$pwd =~ s{/[^/]+$}{};
		$acl = "\n";
		$acl .= "# added by squish (begin)\n";
		# added by squish (begin)
		$acl .= "acl SQUISHLOC dst $hostname";
		$acl .= "acl SQUISHED1 proxy_auth \"/etc/squid/squished\"";
		$acl .= "acl SQUISHED2 ident      \"/etc/squid/squished\"";
		$acl .= "acl SQUISHED3 src        \"/etc/squid/squished\"";
		$acl .= "deny_info http://$hostname/squish/squish.cgi?squished SQUISHED1";
		$acl .= "deny_info http://$hostname/squish/squish.cgi?squished SQUISHED2";
		$acl .= "deny_info http://$hostname/squish/squish.cgi?squished SQUISHED3";
		$acl .= "http_access allow SQUISHLOC";
		$acl .= "http_access deny SQUISHED1";
		$acl .= "http_access deny SQUISHED2";
		$acl .= "http_access deny SQUISHED3";
		$acl .= "# added by squish (end)\n";
		$acl .= "\n";
		$conf =~ s/^(http_access\s+allow)/$acl$1/m;
		print "Writing to $file\n";
		open CONF,"> $file";
		print CONF $conf;
		close CONF;
	}
	# apache include file
	$file="$pwd/apache-squish.conf";
	open CONF,"< $file" || die "$0: can't read $file: $!\n";
	local $/;           # enable "slurp" mode
	$conf=<CONF>;
	close CONF;

	$conf =~ s{\"/.*?\"}{"$pwd/"}g;
	print "Writing to $file\n";
	open CONF,"> $file";
	print CONF $conf;
	close CONF;

	# Bye
	exit(0);
}

while ($ARGV[0] =~ m/^(-.*)/) {
	$option = $1;
	if ($option eq "--exclude") {
		shift @ARGV;
		$excludefile = $ARGV[0] || usage();
		undef @excludelist;
		# Some names to use when anonymising entries
		open EXCLUDEFILE,"< $excludefile" || die "$0: $excludefile: $!\n";
		while (<EXCLUDEFILE>) {
			chomp;
			s/[^\\]#.*//;	# remove unquoted hashes
			s/^#.*//;	# remove unquoted hashes
			s/\\#/#/g;	# unquote quoted hashes
			s/^\s*//;	# remove leading whitespace
			if (m/^$/) { next; }
			if (m/(\S*)\s*(.*?)\s*$/) {
				my $field = $1;
				my $pattern = $2;
				if ($exclude_fieldnames{$field} ne '') {
					$field = $exclude_fieldnames{$field};
				}
				if ($field > 0 || $field eq "0" ) {
					push @excludelist, { "field" => $field, "pattern" => $pattern };
					# print STDERR "$field: '$pattern'\n";
				}
				else {
					print STDERR "Warning: Ignored exclusion \"$_\"\n";
					foreach $b (keys %exclude_fieldnames) {
						print "$b: ".$exclude_fieldnames{$b}."\n";
					}
				}
			}
			else {
				print STDERR "Warning: Ignored exclusion \"$_\"\n";
			}
		}
		close EXCLUDEFILE;
	}
	elsif ($option eq "--nodns") {
		$nodns = 1;
	}
	elsif ($option eq "--config") {
		shift @ARGV;
		$conffile = $ARGV[0] || usage();
	}
	elsif ($option eq "--squish") {
		shift @ARGV;
		$blacklistfile = $ARGV[0] || usage();
	}
	elsif ($option eq "--install") {
		installit();
	}
	elsif ($option eq "--notime") {
		$notime=1;
	}
	else {
		die "$0: Unrecognised option $option\n";
	}
	shift @ARGV;
}

$hostname = $ENV{HOSTNAME} or $ENV{HTTP_HOST};
if ($useanonnames) {
	$hostname = 'squid.example.com';
	$revision = 'squint.pl - http://www.ledge.co.za/software/squish/';
}
# $chdir  = shift @ARGV || usage();
$timefrom = shift @ARGV || 0;
$timeto  = shift @ARGV || 0;

# activesites{$user}{$site} = $lastvisited;
# Determine the period - if we know it
if ($timefrom) {
	$timefromascii = strftime ("%a %d %b %Y", localtime($timefrom));
	$timetoascii = strftime ("%a %d %b %Y", localtime($timeto-1));
	$reportperiod = "$timefromascii to $timetoascii";
}
else {
	$reportperiod = "";
}


# $timezero = `date +\%s -d '2 January 1972'`; # hacky FIXME to get timezone
#$timezero = POSIX::mktime(0, 0, 0, 2, 0, 1972-1900); # 0==january


# Seconds that must pass before an on-line session is considered terminated
$onlinetimeout = 5*60;	# session timeout in seconds - if you download something every five minutes, you're on all day.
			# Lazy bums.  Do some work, and stop wasting time
			# hammering the internet.

my $last_line = \$$state{"last_line"};
my $last_time = \$$state{"last_time"};
my $timefrom = $$state{"last_time"};

if (-t STDIN) {
	usage();
	exit;
}
LOGLINE:
while (<STDIN>) {
	$linecount++;
	@line = split / +/;

	# 0=date 1=transfer-msec? 2=ipaddr 3=status/code 4=size 5=operation
	# 6=url 7=user 8=method/connected-to... 9=content-type
	$time = $line[0];
	$time =~ s/\..*//; # discard milliseconds

	if (defined $timefrom && $_ eq $$last_line) {
		# we're looking at the last line we processed previously
		undef $timefrom;
		next;  # continue with next line -- even if the time is the same
	}
	# Ignore stuff before the starting time
	if (defined $timefrom && $time <= $timefrom) {
		next;
	}
	undef $timefrom;  # don't do that check up there again
	# Remember this line ... we don't want to see it again
	$$last_line=$_;


	# Get and normalise user
	$user = lc($line[7]);
	# $line[3] =~ m/\b(NONE|TCP_DENIED)\b/ and next;	# exclude failed whatevers
	if ( $user eq "-" ) {
		$user = $line[2];
		if ($nodns) {
			$user =~ m/^[\d.]+$/ or next;	# only valid ip address things, or reject the record
		}
	}
	# Remember the IP address which it came from ...
	$ip=$line[2];

	$transferseconds = int(($line[1]+500)/1000); # round off -discard fraction
	# track usage patterns - time of day and day of week:
	#	$hour = int(($time-$timezero) / 3600) % 24;
        #   0    1    2     3     4    5     6     7     8
        #($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) =

	my @tm = localtime($time);
	$hour = $tm[2];
	# FIXME: use >1h periods -- 0-3,4-7,7,...14,15,16,17-19,19-24

	# $day  = int(($time-$timezero) / 86400 ) * 86400 + $timezero; # which day was this on ...
	$day = POSIX::mktime(0, 0, 0, @tm[3..8]);

	# check the exclude list and exclude if we match
	foreach $rule (@excludelist) {
		if ($line[$$rule{"field"}] =~ m/$$rule{"pattern"}/) {
			$$rule{"count"}++;
			$$rule{"size"} += $transferbytes; # byte count
			$$rule{"user"}{$user}++;
			$$rule{"ip"}{$ip}++;

			$excludetotal{"count"}++;
			$excludetotal{"size"} += $transferbytes; # byte count
			$excludetotal{"user"}{$user}++;
			$excludetotal{"ip"}{$ip}++;
			next LOGLINE;
		}
	}
	
	# Get site name from URL:
	if ( $line[6] =~ m{[^:]+://([^:@]+(:[^@]+)?\@)?([^:/]+)(:\d+)?/} )  {
		$url = $line[6];
		$site = $3; # FIXME we should not silently discard protocol and port
	}
	elsif ( $line[5] eq 'CONNECT' ) { # CONNECT URL's are site name and port only
		# FIXME: we should probably count this as one `page' or so ...
		$site = $line[6];
		$url = "https://$site/";  # we're guessing
		$site =~ s/:\d*//; # Strip :443
		#$userpages{$user}++;
		#$pagesviewed{$user}{$day}{$site}++;
	}
	else {
		# print STDOUT "no match: $line[6]\n";
		next;   # Strange -no site name in url...
	}
	$contenttype = $line[9];
	$transferbytes = $line[4];
	if ($transferbytes == 0) { next ; }

	# We have:
	#	$user (some id for the user)
	#	$transferbytes (bytes)
	#	$time (timestamp as an int)


	# guesstimate time on-line -- we're going to work like this:
	# When a download terminates, it is logged.
	# First guess is the transfer time
	# If last download was within the $onlinetimeout then use that instead
	$last = $$userdb{$user}{"last_time"};
	$elapsed = $time - $last;
	if ($elapsed < 0) { $elapsed = 0; }
	my $busyseconds = $transferseconds;
	if ($last!=0 && $elapsed <= $onlinetimeout && $busyseconds < $elapsed ) {
		# print "$busyseconds->$elapsed ";
		# print "Using $elapsed instead of $busyseconds\n";
		$busyseconds = $elapsed;
		# This is correct for many small fast downloads, but doesn't
		# really understand long downloads. Ag well.  The rrdsquish will
		# take care of it.
	}
	$$userdb{$user}{"last_time"} = $time;
	my $rrd_data = userrrd($user,"data");
	my $rrd_time = userrrd($user,"time");
	$rrd_time->moveto($time);
	$rrd_data->moveto($time);
	$rrd_data->setvalues($transferbytes,0,$transferseconds);
	if ($busyseconds != 0) {
		$rrd_time->setvalues($busyseconds,0,$busyseconds);
	#	print "b=$busyseconds:".
	#	"".$rrd_time->sum(0,$busyseconds).":".
	#	"".$rrd_time->sum(0,$busyseconds-1)."\n";
	}
	#showuser($user);
	#print "::::::::::::::::::: $user ::::::::::::::::::::::::";
	#$rrd_time->print(0,13000000);
#	print $_;

	$$last_time=$time;
	$lines_squished ++;
}

# adjustbalances(time());

# FIXME: this is only for debugging with old data:
if ($notime != 1) {
	$time=time();
}
# Update the totals -- we'll use this later ...
for $user (keys %$userdb) {
	my $rrd_data = userrrd($user,"data");
	my $rrd_time  = userrrd($user,"time");
	$rrd_time->moveto($time);
	$rrd_data->moveto($time);
	foreach my $i (0,1,2,3) {
		# $$userdb{$user}{"time"} is the rrd itself ...
		$$userdb{$user}{"t:time"}[$i] = $rrd_time->total($i);
		$$userdb{$user}{"t:data"}[$i] = $rrd_data->total($i);
	}
}


# Zap
writekillfile();
# Save
putcurrentstate();

if (-t STDOUT) {
	showdata();
}
print STDOUT "$0: $lines_squished lines squished\n";
